import logging
import hashlib
from org_mapper.kafka_util import KafkaManager
from org_mapper.storage_manager import StorageManager

_logger = logging.getLogger(__name__)


class KafkaTopicCleanup:
    def __init__(self, docbuilder_overrides_key, config):
        self.docbuilder_overrides_key = docbuilder_overrides_key
        self.kafka_manager = KafkaManager(config)
        self.storage_manager = StorageManager()
        self.admin_bucket = config.get("infra", "admin_bucket")
        self.topic_watermark_cache_key = config.get("kafka", "topic_watermark_cache_key")
        self.max_topics_delete = config.getint("kafka", 'max_topics_delete')
        self.dry_run = config.getboolean("kafka", 'dry_run')

    def clean_topics(self):
        cached_topic_offsets = self.storage_manager.get_json_object(self.topic_watermark_cache_key, self.admin_bucket)
        _logger.info(f"We have {len(cached_topic_offsets) if cached_topic_offsets else 'MISSING_CACHE'} "
                     f"topic offsets in cache")
        new_topic_offsets = self.kafka_manager.get_offsets_all_consumer_groups()
        _logger.info(f"Kafka returned {len(new_topic_offsets)} topic offsets")
        used_topics = self._generate_all_used_topic_hashes()
        _logger.info(f"We have {len(used_topics) if used_topics else 'MISSING_USED'} "
                     f"used topics that have hashes generated")
        if not used_topics:
            _logger.warning("No existing doc builder overrides file, will not clean any topics")
            topics_to_delete = []
        elif not cached_topic_offsets:
            _logger.warning("No existing cache file found, saving one now, will not clean any topics")
            topics_to_delete = []
        else:
            topics_to_delete = self._find_topics_to_delete(cached_topic_offsets, new_topic_offsets, used_topics)
            if not self.dry_run:
                # only delete the first x topics, leave the rest for the next day.
                if len(topics_to_delete) > self.max_topics_delete:
                    _logger.warning(f"Not deleting {len(topics_to_delete) - self.max_topics_delete} topics due to "
                                    f"topics to delete greater than configured max of {self.max_topics_delete}")
                topics_to_delete = topics_to_delete[:self.max_topics_delete]
                self.kafka_manager.delete_topics(topics_to_delete)
            else:
                _logger.warning(f"Dry run mode, We have {len(topics_to_delete)} topics to delete, "
                                f"Max Topics to delete is {self.max_topics_delete}")
        self.storage_manager.upload_json_object(new_topic_offsets, self.topic_watermark_cache_key,
                                                self.admin_bucket)
        return len(topics_to_delete)

    @staticmethod
    def _find_topics_to_delete(cached_topic_offsets, new_topic_offsets, used_topics):
        topics_to_delete = []
        for topic_name in new_topic_offsets:
            # First check all consumer groups of the topic are at the same watermark
            consumer_group_offsets = new_topic_offsets[topic_name]
            _logger.debug(f"Consumer Group Offsets {consumer_group_offsets}")
            result = True
            first_val = list(consumer_group_offsets.values())[0]
            for consumer_group in consumer_group_offsets:
                if consumer_group_offsets[consumer_group] != first_val:
                    result = False
                    _logger.debug(f"Topic {topic_name} has different offsets for current consumers")
                    break
            #  If they are all equal,
            #  go on to check the cache if all cached consumer groups are equal to the new watermarks
            if result is True:
                _logger.debug(f"Topic {topic_name} has same offsets for current consumers")
                cache_result = True
                try:
                    cached_consumer_group_offsets = cached_topic_offsets[topic_name]
                except KeyError:
                    _logger.debug(f"Topic {topic_name} does not exist in the cache")
                    cache_result = False
                else:
                    _logger.debug(f"Cache Consumer Group Offsets {cached_consumer_group_offsets}")
                    for consumer_group in cached_consumer_group_offsets:
                        if cached_consumer_group_offsets[consumer_group] != first_val:
                            cache_result = False
                            _logger.debug(f"Topic {topic_name} has different offsets in cache")
                            break
                # add any topics where the watermarks are all the same across all consumers and with the cached values
                # and the topic isn't currently in the used_topics.
                if cache_result is True:
                    _logger.debug(f"Topic {topic_name} has same offsets in cache")
                if cache_result is True and topic_name not in used_topics:
                    _logger.info(f"Topic {topic_name} is not in used_topics flagging for deletion")
                    topics_to_delete.append(topic_name)
        return topics_to_delete

    @staticmethod
    def convert_topic_names_topic_hash(key, size):
        if size == 0:
            size = 1
        tn = key.replace(':', '-')
        hashes = list()
        mapping = dict()
        for n in range(size):
            org_hash = 'd' + hashlib.md5((tn + '-' + str(n)).encode('utf-8'),
                                         usedforsecurity=False).hexdigest()[0:10].upper()
            hashes.append(org_hash)
            mapping[tn + '-' + str(n)] = org_hash
        return hashes, mapping

    def _generate_all_used_topic_hashes(self):
        docbuilder_overrides = self.storage_manager.get_json_object(self.docbuilder_overrides_key, self.admin_bucket)
        in_use_topic_hashes = []
        if docbuilder_overrides:
            for org in docbuilder_overrides:
                hashes, mappings = self.convert_topic_names_topic_hash(org, docbuilder_overrides[org])
                in_use_topic_hashes += hashes
            return set(in_use_topic_hashes)
        else:
            return None
