import os
import logging
from datetime import datetime
from threading import Thread

import boto3
import schedule
import time
from pscr_common.log import setup_logging, clear_logging_handlers
from pscr_common.health_check import update_health_check

from org_mapper import timestream_util
from org_mapper.app_metrics import Metrics
from org_mapper.metrics_v2 import TimeStreamMetrics

from org_mapper.lc_transformer import LCTransformer
from org_mapper.nta_transformer import NTATransformer
from org_mapper.utils import Config, str_to_bool
from org_mapper.report_identifier_metrics import IdentifierStatsReporter
from org_mapper.storage_manager import StorageManager
from org_mapper.kafka_topic_manager import KafkaTopicCleanup

logging.getLogger("boto").setLevel(logging.CRITICAL)
logging.getLogger("botocore").setLevel(logging.CRITICAL)
logging.getLogger("s3transfer").setLevel(logging.CRITICAL)
logging.getLogger("urllib3").setLevel(logging.CRITICAL)
_logger = logging.getLogger(__name__)


class OrgMapper:
    region = boto3.session.Session().region_name

    def __init__(self, metrics_obj=None):
        self.metrics = metrics_obj.metrics_registry
        self.period_back_metric_points = None
        self.config = Config.create()
        self.shards = None
        self.env_id = self.config.get("env", "id")
        self.shard_map_key_prefix = self.config.get(
            "shard_management", "shard_map_key_prefix"
        )
        self.output_shard_map = self.config.getboolean(
            "shard_management", "output_shard_map"
        )
        self.disable_shard_merge = self.config.getboolean(
            "shard_management", "disable_shard_merge"
        )
        self.disable_shard_split = self.config.getboolean(
            "shard_management", "disable_shard_split"
        )
        self.dry_run = self.config.getboolean("shard_management", "dry_run")
        self.interval_minutes = self.config.getint("service", "interval_minutes")
        self.top_x_points = self.config.getint("metrics", "top_x_points")
        self.last_x_points = self.config.getint("metrics", "last_x_points")
        self.save_effective_endpoint_cache = self.config.getboolean(
            "lc_effective_endpoints", "save_effective_endpoint_cache"
        )
        self.last_hour_higher_factor = self.config.getfloat(
            "lc_effective_endpoints", "last_hour_higher_factor"
        )
        self.cache_refresh_interval_min = self.config.getint(
            "timestream", "cache_refresh_interval_min"
        )
        self.timestream_enabled = str_to_bool(self.config.get("timestream", "enable"))
        self.timestream_supported_regions = [
            region.strip()
            for region in self.config.get("timestream", "supported_regions").split(",")
        ]
        self.timestream_cache_enabled = str_to_bool(
            self.config.get("timestream", "cache")
        )
        self.mongodb_enabled = str_to_bool(self.config.get("mongo", "enable"))
        self.period_back_metric_points = self.config.getint(
            "metrics", "data_rate_period_back_metric_points"
        )
        mode = self.config.get("service", "mode")
        if mode == "LC":
            _logger.info(f"Mode {mode}")
            self.transformer_class = LCTransformer
            self.docbuilder_overrides_key = self.config.get("lc_effective_endpoints", "doc_builder_override_key")
        elif mode == "NTA":
            _logger.info(f"Mode {mode}")
            self.transformer_class = NTATransformer
            self.docbuilder_overrides_key = self.config.get("nta_effective_bandwidth", "doc_builder_override_key")
        else:
            _logger.error(f"Undefined Mode {mode}")
            self.transformer_class = None
        self.topic_cleanup_enable = self.config.getboolean("kafka", "enable")
        self.topic_cleanup_time = self.config.get("kafka", "time_run_topic_deletion_gmt")
        self.topic_run_hourly = self.config.getboolean("kafka", "run_hourly")
        self.admin_bucket = self.config.get("infra", "admin_bucket")
        self.storage_manager = StorageManager()
        self.partition_map_key_prefix = self.config.get("kafka", "partition_map_key_prefix")

    def calculate_partitions_count(self):
        with self.metrics.timer("calculate_partitions_count").time():
            start_time = datetime.utcfromtimestamp(time.time())
            if self.mongodb_enabled:
                _logger.info(f"Pulling Metrics from MongoDB")
                transformer = self.transformer_class(
                    self.period_back_metric_points, "MongoDB", self.config
                )
            elif (
                self.timestream_enabled
                and OrgMapper.region in self.timestream_supported_regions
            ):
                _logger.info(f"Pulling Metrics from Timestream")
                transformer = self.transformer_class(
                    self.period_back_metric_points, "TimeStream", self.config
                )
            else:
                _logger.info(f"Pulling Metrics from CloudWatch")
                transformer = self.transformer_class(
                    self.period_back_metric_points, "CloudWatch", self.config
                )

            update_health_check()
            transformer.convert_bandwidth_shard_map()
            update_health_check()

            partition_keys = transformer.partition_keys
            partition_metadata = transformer.partition_metadata

            transformer.flush_docbuilder_overrides()
            self.output_partition_map()

            _logger.info(
                f"Total Time: {datetime.utcfromtimestamp(time.time()) - start_time}"
            )

    def output_partition_map(self):
        docbuilder_overrides = self.storage_manager.get_json_object(self.docbuilder_overrides_key, self.admin_bucket)
        topic_mappings = dict()
        if docbuilder_overrides:
            for org in docbuilder_overrides:
                hashes, mappings = KafkaTopicCleanup.convert_topic_names_topic_hash(org, docbuilder_overrides[org])
                topic_mappings.update(mappings)
            timestamp = datetime.now().isoformat()
            self.storage_manager.upload_json_object(topic_mappings, f"{self.partition_map_key_prefix}latest.txt",
                                                    self.admin_bucket)
            self.storage_manager.upload_json_object(topic_mappings, f"{self.partition_map_key_prefix}{timestamp}.txt",
                                                    self.admin_bucket)

    def org_mapper_run(self):
        _logger.info(
            f"Starting up using the following: period_back_metric_points: {self.period_back_metric_points}"
        )
        self.calculate_partitions_count()
        timestream_util.prev_run_time = TimeStreamMetrics.prev_run_time

    def kafka_topic_cleanup_run(self):
        _logger.info("Starting Kafka manager")
        kafka_cleanup = KafkaTopicCleanup(self.docbuilder_overrides_key, self.config)
        kafka_cleanup.clean_topics()

    def schedule_run(self):
        scheduler_main = schedule.Scheduler()
        scheduler_main.every(self.interval_minutes).minutes.do(self.org_mapper_run)
        if self.topic_cleanup_enable:
            # This scheduler library is simplistic and runs any pending tasks serially, which is actually what we need
            # in this case we never want org mapper main loop and the topic cleanup to run in parallel.
            # Topic cleanup is meant to run once per day. We also have an override to run hourly.
            if self.topic_run_hourly:
                scheduler_main.every(60).minutes.do(self.kafka_topic_cleanup_run)
            else:
                scheduler_main.every().day.at(self.topic_cleanup_time).do(self.kafka_topic_cleanup_run)

        while True:
            scheduler_main.run_pending()
            time.sleep(1)

    def run_once(self):
        self.org_mapper_run()
        if self.topic_cleanup_enable:
            self.kafka_topic_cleanup_run()


def main():
    clear_logging_handlers()
    if str_to_bool(os.environ.get("PSCR_LOGGING", "true")):
        setup_logging(
            level=os.environ.get("LOG_LEVEL", "INFO"),
            svc_version=os.environ.get("CONTAINERVERSION", "UNKNOWN"),
            env_id=os.environ.get("ENV_ID", "UNKNOWN"),
            svc_id="TODO",
        )
    else:
        logging.basicConfig(level=os.environ.get("LOG_LEVEL", "DEBUG"))

    metrics_obj = Metrics()
    metrics_obj.start_reporter()

    if str_to_bool(os.environ.get("ENABLE_IDENTIFIER_METRIC_REPORTING", "false")):
        org_stats_reporter = IdentifierStatsReporter(Config.create(), metrics_obj)
        p = Thread(target=org_stats_reporter.schedule_run)
        p.start()
    update_health_check()
    run_once = str_to_bool(os.environ.get("RUN_ONCE", "false"))

    org_mapper = OrgMapper(metrics_obj)
    if run_once:
        _logger.info("Starting org mapper in Run Once Mode")
        org_mapper.run_once()
    else:
        _logger.info("Starting org mapper in Schedule Mode")
        org_mapper.schedule_run()


if __name__ == "__main__":
    main()
